@@ -24,6 +24,7 @@
 #include <string.h>
 #include <signal.h>
 #include <stdlib.h>
+#include <errno.h>
 #include <dbus/dbus.h>
 #include "gdaemonvfs.h"
 #include "gvfsuriutils.h"
@@ -36,7 +37,10 @@
 #include "gmountspec.h"
 #include "gvfsurimapper.h"
 #include "gdaemonvolumemonitor.h"
+#include "gvfsicon.h"
+#include "gvfsiconloadable.h"
 #include <glib/gi18n-lib.h>
+#include <glib/gstdio.h>
 
 typedef struct  {
   char *type;
@@ -55,6 +59,8 @@ struct _GDaemonVfs
   GVfs *wrapped_vfs;
   GList *mount_cache;
 
+  GFile *fuse_root;
+  
   GHashTable *from_uri_hash;
   GHashTable *to_uri_hash;
 
@@ -230,7 +236,7 @@ get_mountspec_from_uri (GDaemonVfs *vfs,
 	      if (mountable && mountable->host_is_inet)
 		{
 		  /* Convert hostname to lower case */
-          str_tolower_inplace (decoded->host);
+		  str_tolower_inplace (decoded->host);
 		  
 		  /* Remove brackets aroung ipv6 addresses */
 		  l = strlen (decoded->host);
@@ -287,6 +293,7 @@ g_daemon_vfs_init (GDaemonVfs *vfs)
   const char * const *schemes, * const *mount_types;
   GVfsUriMapper *mapper;
   GList *modules;
+  char *file;
   int i;
 
   bindtextdomain (GETTEXT_PACKAGE, GVFS_LOCALEDIR);
@@ -319,6 +326,10 @@ g_daemon_vfs_init (GDaemonVfs *vfs)
   
   vfs->wrapped_vfs = g_vfs_get_local ();
 
+  file = g_build_filename (g_get_home_dir(), ".gvfs", NULL);
+  vfs->fuse_root = g_vfs_get_file_for_path (vfs->wrapped_vfs, file);
+  g_free (file);
+  
   dbus_connection_set_exit_on_disconnect (vfs->async_bus, FALSE);
 
   _g_dbus_connection_integrate_with_main (vfs->async_bus);
@@ -357,13 +368,42 @@ g_daemon_vfs_new (void)
   return g_object_new (G_TYPE_DAEMON_VFS, NULL);
 }
 
+/* This unrefs file if its changed */
+static GFile *
+convert_fuse_path (GVfs     *vfs,
+		   GFile    *file)
+{
+  GFile *fuse_root;
+  char *fuse_path, *mount_path;
+  GMountInfo *mount_info;
+
+  fuse_root = ((GDaemonVfs *)vfs)->fuse_root;
+  if (g_file_has_prefix (file, fuse_root))
+    {
+      fuse_path = g_file_get_path (file);
+      mount_info = _g_daemon_vfs_get_mount_info_by_fuse_sync (fuse_path, &mount_path);
+      g_free (fuse_path);
+      if (mount_info)
+	{
+	  g_object_unref (file);
+	  /* TODO: Do we need to look at the prefix of the mount_spec? */
+	  file = g_daemon_file_new (mount_info->mount_spec, mount_path);
+	  g_free (mount_path);
+	  g_mount_info_unref (mount_info);
+	}
+    }
+  return file;
+}
+
 static GFile *
 g_daemon_vfs_get_file_for_path (GVfs       *vfs,
 				const char *path)
 {
-  /* TODO: detect fuse paths and convert to daemon vfs GFiles */
+  GFile *file;
   
-  return g_vfs_get_file_for_path (G_DAEMON_VFS (vfs)->wrapped_vfs, path);
+  file = g_vfs_get_file_for_path (G_DAEMON_VFS (vfs)->wrapped_vfs, path);
+  file = convert_fuse_path (vfs, file);
+  return file;
 }
 
 static GFile *
@@ -711,6 +751,41 @@ lookup_mount_info_in_cache (GMountSpec *spec,
   return info;
 }
 
+static GMountInfo *
+lookup_mount_info_by_fuse_path_in_cache (const char *fuse_path,
+					 char **mount_path)
+{
+  GMountInfo *info;
+  GList *l;
+
+  G_LOCK (mount_cache);
+  info = NULL;
+  for (l = the_vfs->mount_cache; l != NULL; l = l->next)
+    {
+      GMountInfo *mount_info = l->data;
+
+      if (mount_info->fuse_mountpoint != NULL &&
+	  g_str_has_prefix (fuse_path, mount_info->fuse_mountpoint))
+	{
+	  int len = strlen (mount_info->fuse_mountpoint);
+	  if (fuse_path[len] == 0 ||
+	      fuse_path[len] == '/')
+	    {
+	      if (fuse_path[len] == 0)
+		*mount_path = g_strdup ("/");
+	      else
+		*mount_path = g_strdup (fuse_path + len);
+	      info = g_mount_info_ref (mount_info);
+	      break;
+	    }
+	}
+    }
+  G_UNLOCK (mount_cache);
+
+  return info;
+}
+
+
 void
 _g_daemon_vfs_invalidate_dbus_id (const char *dbus_id)
 {
@@ -908,6 +983,71 @@ _g_daemon_vfs_get_mount_info_sync (GMountSpec *spec,
   return info;
 }
 
+GMountInfo *
+_g_daemon_vfs_get_mount_info_by_fuse_sync (const char *fuse_path,
+					   char **mount_path)
+{
+  GMountInfo *info;
+  DBusConnection *conn;
+  DBusMessage *message, *reply;
+  DBusMessageIter iter;
+  DBusError derror;
+  int len;
+	
+  info = lookup_mount_info_by_fuse_path_in_cache (fuse_path,
+						  mount_path);
+  if (info != NULL)
+    return info;
+  
+  conn = _g_dbus_connection_get_sync (NULL, NULL);
+  if (conn == NULL)
+    return NULL;
+
+  message =
+    dbus_message_new_method_call (G_VFS_DBUS_DAEMON_NAME,
+				  G_VFS_DBUS_MOUNTTRACKER_PATH,
+				  G_VFS_DBUS_MOUNTTRACKER_INTERFACE,
+				  G_VFS_DBUS_MOUNTTRACKER_OP_LOOKUP_MOUNT_BY_FUSE_PATH);
+  dbus_message_set_auto_start (message, TRUE);
+  
+  dbus_message_iter_init_append (message, &iter);
+  _g_dbus_message_iter_append_cstring (&iter, fuse_path);
+
+  dbus_error_init (&derror);
+  reply = dbus_connection_send_with_reply_and_block (conn, message, -1, &derror);
+  dbus_message_unref (message);
+  if (!reply)
+    {
+      dbus_error_free (&derror);
+      return NULL;
+    }
+
+  info = handler_lookup_mount_reply (reply, NULL);
+  dbus_message_unref (reply);
+  
+  if (info)
+    {
+      if (info->fuse_mountpoint)
+	{
+	  len = strlen (info->fuse_mountpoint);
+	  if (fuse_path[len] == 0)
+	    *mount_path = g_strdup ("/");
+	  else
+	    *mount_path = g_strdup (fuse_path + len);
+	}
+      else
+	{
+	  /* This could happen if we race with the gvfs fuse mount
+	   * at startup of gvfsd... */
+	  g_mount_info_unref (info);
+	  info = NULL;
+	}
+    }
+
+  
+  return info;
+}
+
 static GFile *
 g_daemon_vfs_parse_name (GVfs       *vfs,
 			 const char *parse_name)
@@ -917,8 +1057,8 @@ g_daemon_vfs_parse_name (GVfs       *vfs,
   if (g_path_is_absolute (parse_name) ||
       *parse_name == '~')
     {
-      /* TODO: detect fuse paths and convert to daemon vfs GFiles ? */
       file = g_vfs_parse_name (G_DAEMON_VFS (vfs)->wrapped_vfs, parse_name);
+      file = convert_fuse_path (vfs, file);
     }
   else
     {
@@ -928,6 +1068,430 @@ g_daemon_vfs_parse_name (GVfs       *vfs,
   return file;
 }
 
+static gboolean
+enumerate_keys_callback (const char *key,
+			 MetaKeyType type,
+			 gpointer value,
+			 gpointer user_data)
+{
+  GFileInfo  *info = user_data;
+  char *attr;
+
+  attr = g_strconcat ("metadata::", key, NULL);
+
+  if (type == META_KEY_TYPE_STRING)
+    g_file_info_set_attribute_string (info, attr, (char *)value);
+  else
+    g_file_info_set_attribute_stringv (info, attr, (char **)value);
+
+  g_free (attr);
+
+  return TRUE;
+}
+
+static void
+g_daemon_vfs_local_file_add_info (GVfs       *vfs,
+				  const char *filename,
+				  guint64     device,
+				  GFileAttributeMatcher *matcher,
+				  GFileInfo  *info,
+				  GCancellable *cancellable,
+				  gpointer    *extra_data,
+				  GDestroyNotify *extra_data_free)
+{
+  MetaLookupCache *cache;
+  const char *first;
+  char *tree_path;
+  gboolean all;
+  MetaTree *tree;
+
+  /* Filename may or may not be a symlink, but we should not follow it.
+     However, we want to follow symlinks for all parents that have the same
+     device node */
+  all = g_file_attribute_matcher_enumerate_namespace (matcher, "metadata");
+
+  first = NULL;
+  if (!all)
+    {
+      first = g_file_attribute_matcher_enumerate_next (matcher);
+
+      if (first == NULL)
+	return; /* No match */
+    }
+
+  if (*extra_data == NULL)
+    {
+      *extra_data = meta_lookup_cache_new ();
+      *extra_data_free = (GDestroyNotify)meta_lookup_cache_free;
+    }
+  cache = (MetaLookupCache *)*extra_data;
+
+  tree = meta_lookup_cache_lookup_path (cache,
+					filename,
+					device,
+					FALSE,
+					&tree_path);
+
+  if (tree)
+    {
+      meta_tree_enumerate_keys (tree, tree_path,
+				enumerate_keys_callback, info);
+      meta_tree_unref (tree);
+      g_free (tree_path);
+    }
+}
+
+static void
+g_daemon_vfs_add_writable_namespaces (GVfs       *vfs,
+				      GFileAttributeInfoList *list)
+{
+  g_file_attribute_info_list_add (list,
+				  "metadata",
+				  G_FILE_ATTRIBUTE_TYPE_STRING, /* Also STRINGV, but no way express this ... */
+				  G_FILE_ATTRIBUTE_INFO_COPY_WITH_FILE |
+				  G_FILE_ATTRIBUTE_INFO_COPY_WHEN_MOVED);
+}
+
+static void
+send_message_oneway (DBusMessage *message)
+{
+  DBusConnection *connection;
+
+  connection = _g_dbus_connection_get_sync (NULL, NULL);
+  if (connection == NULL)
+    return;
+
+  dbus_connection_send (connection, message, NULL);
+  dbus_connection_flush (connection);
+}
+
+/* Sends a message on the session bus and blocks for the reply,
+   using the thread local connection */
+gboolean
+_g_daemon_vfs_send_message_sync (DBusMessage *message,
+				 GCancellable *cancellable,
+				 GError **error)
+{
+  DBusConnection *connection;
+  DBusError derror;
+  DBusMessage *reply;
+
+  connection = _g_dbus_connection_get_sync (NULL, NULL);
+  if (connection == NULL)
+    {
+      g_set_error (error, G_IO_ERROR,
+		   G_IO_ERROR_FAILED,
+		   _("Error setting file metadata: %s"),
+		   _("Can't contact session bus"));
+      return FALSE;
+    }
+
+  dbus_error_init (&derror);
+  /* TODO: Handle cancellable */
+  reply = dbus_connection_send_with_reply_and_block (connection, message,
+						     G_VFS_DBUS_TIMEOUT_MSECS,
+						     &derror);
+  if (!reply)
+    {
+      _g_error_from_dbus (&derror, error);
+      dbus_error_free (&derror);
+      return FALSE;
+    }
+
+  dbus_message_unref (reply);
+
+  return TRUE;
+}
+
+static gboolean
+strv_equal (char **a, char **b)
+{
+  int i;
+
+  if (g_strv_length (a) != g_strv_length (b))
+    return FALSE;
+
+  for (i = 0; a[i] != NULL; i++)
+    {
+      if (strcmp (a[i], b[i]) != 0)
+	return FALSE;
+    }
+  return TRUE;
+}
+
+/* -1 => error, otherwise number of added items */
+int
+_g_daemon_vfs_append_metadata_for_set (DBusMessage *message,
+				       MetaTree *tree,
+				       const char *path,
+				       const char *attribute,
+				       GFileAttributeType type,
+				       gpointer   value)
+{
+  const char *key;
+  int res;
+
+  key = attribute + strlen ("metadata::");
+
+  res = 0;
+  if (type == G_FILE_ATTRIBUTE_TYPE_STRING)
+    {
+      const char *current;
+      const char *val = (char *)value;
+
+      current = meta_tree_lookup_string (tree, path, key);
+      if (current == NULL || strcmp (current, val) != 0)
+	{
+	  res = 1;
+	  _g_dbus_message_append_args (message,
+				       DBUS_TYPE_STRING, &key,
+				       DBUS_TYPE_STRING, &val,
+				       0);
+	}
+    }
+  else if (type == G_FILE_ATTRIBUTE_TYPE_STRINGV)
+    {
+      char **current;
+      char **val = (char **)value;
+      current = meta_tree_lookup_stringv (tree, path, key);
+      if (current == NULL || !strv_equal (current, val))
+	{
+	  res = 1;
+	  _g_dbus_message_append_args (message,
+				       DBUS_TYPE_STRING, &key,
+				       DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &val, g_strv_length (val),
+				       0);
+	}
+    }
+  else if (type == G_FILE_ATTRIBUTE_TYPE_INVALID)
+    {
+      if (meta_tree_lookup_key_type (tree, path, key) != META_KEY_TYPE_NONE)
+	{
+	  char c = 0;
+	  res = 1;
+	  /* Byte => unset */
+	  _g_dbus_message_append_args (message,
+				       DBUS_TYPE_STRING, &key,
+				       DBUS_TYPE_BYTE, &c,
+				       0);
+	}
+    }
+  else
+    res = -1;
+
+  return res;
+}
+
+static gboolean
+g_daemon_vfs_local_file_set_attributes (GVfs       *vfs,
+					const char *filename,
+					GFileInfo  *info,
+					GFileQueryInfoFlags flags,
+					GCancellable *cancellable,
+					GError    **error)
+{
+  GFileAttributeType type;
+  MetaLookupCache *cache;
+  const char *metatreefile;
+  struct stat statbuf;
+  char **attributes;
+  char *tree_path;
+  MetaTree *tree;
+  int errsv;
+  int i, num_set;
+  DBusMessage *message;
+  gboolean res;
+  int appended;
+  gpointer value;
+
+  res = TRUE;
+  if (g_file_info_has_namespace (info, "metadata"))
+    {
+      attributes = g_file_info_list_attributes (info, "metadata");
+
+      if (g_lstat (filename, &statbuf) != 0)
+	{
+	  errsv = errno;
+	  g_set_error (error, G_IO_ERROR,
+		       g_io_error_from_errno (errsv),
+		       _("Error setting file metadata: %s"),
+		       g_strerror (errsv));
+	  error = NULL; /* Don't set further errors */
+
+	  for (i = 0; attributes[i] != NULL; i++)
+	    g_file_info_set_attribute_status (info, attributes[i],
+					      G_FILE_ATTRIBUTE_STATUS_ERROR_SETTING);
+
+	  res = FALSE;
+	}
+      else
+	{
+	  cache = meta_lookup_cache_new ();
+	  tree = meta_lookup_cache_lookup_path (cache,
+						filename,
+						statbuf.st_dev,
+						FALSE,
+						&tree_path);
+	  message =
+	    dbus_message_new_method_call (G_VFS_DBUS_METADATA_NAME,
+					  G_VFS_DBUS_METADATA_PATH,
+					  G_VFS_DBUS_METADATA_INTERFACE,
+					  G_VFS_DBUS_METADATA_OP_SET);
+	  g_assert (message != NULL);
+	  metatreefile = meta_tree_get_filename (tree);
+	  _g_dbus_message_append_args (message,
+				       G_DBUS_TYPE_CSTRING, &metatreefile,
+				       G_DBUS_TYPE_CSTRING, &tree_path,
+				       0);
+	  meta_lookup_cache_free (cache);
+
+	  num_set = 0;
+	  for (i = 0; attributes[i] != NULL; i++)
+	    {
+	      if (g_file_info_get_attribute_data (info, attributes[i], &type, &value, NULL))
+		{
+		  appended = _g_daemon_vfs_append_metadata_for_set (message,
+								    tree,
+								    tree_path,
+								    attributes[i],
+								    type,
+								    value);
+		  if (appended != -1)
+		    {
+		      num_set += appended;
+		      g_file_info_set_attribute_status (info, attributes[i],
+							G_FILE_ATTRIBUTE_STATUS_SET);
+		    }
+		  else
+		    {
+		      res = FALSE;
+		      g_set_error (error, G_IO_ERROR,
+				   G_IO_ERROR_INVALID_ARGUMENT,
+				   _("Error setting file metadata: %s"),
+				   _("values must be string or list of strings"));
+		      error = NULL; /* Don't set further errors */
+		      g_file_info_set_attribute_status (info, attributes[i],
+							G_FILE_ATTRIBUTE_STATUS_ERROR_SETTING);
+		    }
+		}
+	    }
+
+	  meta_tree_unref (tree);
+	  g_free (tree_path);
+
+	  if (num_set > 0 &&
+	      !_g_daemon_vfs_send_message_sync (message,
+						cancellable, error))
+	    {
+	      res = FALSE;
+	      error = NULL; /* Don't set further errors */
+	      for (i = 0; attributes[i] != NULL; i++)
+		g_file_info_set_attribute_status (info, attributes[i],
+						  G_FILE_ATTRIBUTE_STATUS_ERROR_SETTING);
+	    }
+
+	  dbus_message_unref (message);
+	}
+
+      g_strfreev (attributes);
+    }
+
+  return res;
+}
+
+static void
+g_daemon_vfs_local_file_removed (GVfs       *vfs,
+				 const char *filename)
+{
+  MetaLookupCache *cache;
+  DBusMessage *message;
+  const char *metatreefile;
+  MetaTree *tree;
+  char *tree_path;
+
+  cache = meta_lookup_cache_new ();
+  tree = meta_lookup_cache_lookup_path (cache,
+					filename,
+					0,
+					FALSE,
+					&tree_path);
+  if (tree)
+    {
+      message =
+	dbus_message_new_method_call (G_VFS_DBUS_METADATA_NAME,
+				      G_VFS_DBUS_METADATA_PATH,
+				      G_VFS_DBUS_METADATA_INTERFACE,
+				      G_VFS_DBUS_METADATA_OP_REMOVE);
+      g_assert (message != NULL);
+      metatreefile = meta_tree_get_filename (tree);
+      _g_dbus_message_append_args (message,
+				   G_DBUS_TYPE_CSTRING, &metatreefile,
+				   G_DBUS_TYPE_CSTRING, &tree_path,
+				   0);
+      send_message_oneway (message);
+      dbus_message_unref (message);
+      meta_tree_unref (tree);
+      g_free (tree_path);
+    }
+
+  meta_lookup_cache_free (cache);
+}
+
+static void
+g_daemon_vfs_local_file_moved (GVfs       *vfs,
+			       const char *source,
+			       const char *dest)
+{
+  MetaLookupCache *cache;
+  DBusMessage *message;
+  const char *metatreefile;
+  MetaTree *tree1, *tree2;
+  char *tree_path1, *tree_path2;
+
+  cache = meta_lookup_cache_new ();
+  tree1 = meta_lookup_cache_lookup_path (cache,
+					 source,
+					 0,
+					 FALSE,
+					 &tree_path1);
+  tree2 = meta_lookup_cache_lookup_path (cache,
+					 dest,
+					 0,
+					 FALSE,
+					 &tree_path2);
+  if (tree1 && tree2 && tree1 == tree2)
+    {
+      message =
+	dbus_message_new_method_call (G_VFS_DBUS_METADATA_NAME,
+				      G_VFS_DBUS_METADATA_PATH,
+				      G_VFS_DBUS_METADATA_INTERFACE,
+				      G_VFS_DBUS_METADATA_OP_MOVE);
+      g_assert (message != NULL);
+      metatreefile = meta_tree_get_filename (tree1);
+      _g_dbus_message_append_args (message,
+				   G_DBUS_TYPE_CSTRING, &metatreefile,
+				   G_DBUS_TYPE_CSTRING, &tree_path1,
+				   G_DBUS_TYPE_CSTRING, &tree_path2,
+				   0);
+      send_message_oneway (message);
+      dbus_message_unref (message);
+    }
+
+  if (tree1)
+    {
+      meta_tree_unref (tree1);
+      g_free (tree_path1);
+    }
+
+  if (tree2)
+    {
+      meta_tree_unref (tree2);
+      g_free (tree_path2);
+    }
+
+  meta_lookup_cache_free (cache);
+}
+
 DBusConnection *
 _g_daemon_vfs_get_async_bus (void)
 {
@@ -951,11 +1515,11 @@ g_daemon_vfs_class_init (GDaemonVfsClass *class)
 {
   GObjectClass *object_class;
   GVfsClass *vfs_class;
-  
+
   object_class = (GObjectClass *) class;
 
   g_daemon_vfs_parent_class = g_type_class_peek_parent (class);
-  
+
   object_class->finalize = g_daemon_vfs_finalize;
 
   vfs_class = G_VFS_CLASS (class);
@@ -965,6 +1529,11 @@ g_daemon_vfs_class_init (GDaemonVfsClass *class)
   vfs_class->get_file_for_uri = g_daemon_vfs_get_file_for_uri;
   vfs_class->get_supported_uri_schemes = g_daemon_vfs_get_supported_uri_schemes;
   vfs_class->parse_name = g_daemon_vfs_parse_name;
+  vfs_class->local_file_add_info = g_daemon_vfs_local_file_add_info;
+  vfs_class->add_writable_namespaces = g_daemon_vfs_add_writable_namespaces;
+  vfs_class->local_file_set_attributes = g_daemon_vfs_local_file_set_attributes;
+  vfs_class->local_file_removed = g_daemon_vfs_local_file_removed;
+  vfs_class->local_file_moved = g_daemon_vfs_local_file_moved;
 }
 
 /* Module API */
@@ -983,8 +1552,20 @@ g_io_module_load (GIOModule *module)
   if (g_getenv ("DBUS_SESSION_BUS_ADDRESS") == NULL) 
     return;
 
+  /* Make this module resident so that we ground the common
+   * library. If that is unloaded we could get into all kinds
+   * of strange situations. This is safe to do even if we loaded
+   * some other common-using module first as all modules are loaded
+   * before any are freed.
+   */
+  g_type_module_use (G_TYPE_MODULE (module));
+  
   g_daemon_vfs_register_type (G_TYPE_MODULE (module));
   g_daemon_volume_monitor_register_types (G_TYPE_MODULE (module));
+  
+  /* We implement GLoadableIcon only on client side.
+     see comment in common/giconvfs.c */
+  _g_vfs_icon_add_loadable_interface ();
 
   g_io_extension_point_implement (G_VFS_EXTENSION_POINT_NAME,
 				  G_TYPE_DAEMON_VFS,
