@@ -29,7 +29,6 @@
 #include <unistd.h>
 #include <stdio.h>
 #include <errno.h>
-#include <poll.h>
 
 #include <glib/gi18n-lib.h>
 
@@ -301,7 +300,7 @@ setup_async_fd_receive (VfsConnectionData *connection_data)
   
   
   connection_data->extra_fd_source =
-    __g_fd_source_new (connection_data->extra_fd, POLLIN|POLLERR, NULL);
+    __g_fd_source_new (connection_data->extra_fd, G_IO_IN|G_IO_ERR, NULL);
   g_source_set_callback (connection_data->extra_fd_source,
 			 (GSourceFunc)async_connection_accept_new_fd,
 			 connection_data, NULL);
@@ -509,10 +508,19 @@ async_call_send (AsyncDBusCall *async_call)
 {
   AsyncCallCancelData *cancel_data;
 
+  _g_dbus_connection_call_async (async_call->connection,
+				 async_call->message,
+				 G_VFS_DBUS_TIMEOUT_MSECS,
+				 async_dbus_response,
+				 async_call);
+
   if (async_call->cancellable)
     {
       cancel_data = g_new0 (AsyncCallCancelData, 1);
       cancel_data->connection = dbus_connection_ref (async_call->connection);
+      /* make sure we get the serial *after* the message has been sent, otherwise
+       * it will be 0
+       */
       cancel_data->serial = dbus_message_get_serial (async_call->message);
       async_call->cancelled_tag =
 	g_signal_connect_data (async_call->cancellable, "cancelled",
@@ -521,12 +529,6 @@ async_call_send (AsyncDBusCall *async_call)
 			       (GClosureNotify)async_call_cancel_data_free,
 			       0);
     }
-
-  _g_dbus_connection_call_async (async_call->connection,
-				 async_call->message,
-				 G_VFS_DBUS_TIMEOUT_MSECS,
-				 async_dbus_response,
-				 async_call);
 }
 
 static void
@@ -745,16 +747,16 @@ _g_vfs_daemon_call_sync (DBusMessage *message,
       sent_cancel = (cancel_fd == -1);
       while (!dbus_pending_call_get_completed (pending))
 	{
-	  struct pollfd poll_fds[2];
+	  GPollFD poll_fds[2];
 	  int poll_ret;
 	  
 	  do
 	    {
-	      poll_fds[0].events = POLLIN;
+	      poll_fds[0].events = G_IO_IN;
 	      poll_fds[0].fd = dbus_fd;
-	      poll_fds[1].events = POLLIN;
+	      poll_fds[1].events = G_IO_IN;
 	      poll_fds[1].fd = cancel_fd;
-	      poll_ret = poll (poll_fds, sent_cancel?1:2, -1);
+	      poll_ret = g_poll (poll_fds, sent_cancel?1:2, -1);
 	    }
 	  while (poll_ret == -1 && errno == EINTR);
 
@@ -881,10 +883,10 @@ invalidate_local_connection (const char *dbus_id,
   if (local)
     g_hash_table_remove (local->connections, dbus_id);
   
-  g_set_error (error,
-	       G_VFS_ERROR,
-	       G_VFS_ERROR_RETRY,
-	       "Cache invalid, retry (internally handled)");
+  g_set_error_literal (error,
+		       G_VFS_ERROR,
+		       G_VFS_ERROR_RETRY,
+		       "Cache invalid, retry (internally handled)");
 }
 
 DBusConnection *
@@ -954,7 +956,7 @@ _g_dbus_connection_get_sync (const char *dbus_id,
       if (bus == NULL)
 	{
 	  g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-		       "Couldn't get main dbus connection: %s\n",
+		       "Couldn't get main dbus connection: %s",
 		       derror.message);
 	  dbus_error_free (&derror);
 	  return NULL;
@@ -1022,3 +1024,28 @@ _g_dbus_connection_get_sync (const char *dbus_id,
 
   return connection;
 }
+
+/**
+ * _g_simple_async_result_complete_with_cancellable:
+ * @result: the result
+ * @cancellable: a cancellable to check
+ *
+ * If @cancellable is cancelled, sets @result into the cancelled error
+ * state. Then calls g_simple_async_result_complete().
+ * This function is useful to ensure that @result is properly set into
+ * an error state on cancellation.
+ **/
+void
+_g_simple_async_result_complete_with_cancellable (GSimpleAsyncResult *result,
+                                                  GCancellable       *cancellable)
+{
+  if (cancellable &&
+      g_cancellable_is_cancelled (cancellable))
+    g_simple_async_result_set_error (result,
+                                     G_IO_ERROR,
+                                     G_IO_ERROR_CANCELLED,
+                                     "%s", _("Operation was cancelled"));
+
+  g_simple_async_result_complete (result);
+}
+
