@@ -35,6 +35,7 @@
 #include "gvfsjobsource.h"
 #include "gvfsdaemonprotocol.h"
 #include <gvfsjobopenforread.h>
+#include <gvfsjobopeniconforread.h>
 #include <gvfsjobopenforwrite.h>
 #include <gvfsjobqueryinfo.h>
 #include <gvfsjobqueryfsinfo.h>
@@ -45,12 +46,16 @@
 #include <gvfsjobunmount.h>
 #include <gvfsjobmountmountable.h>
 #include <gvfsjobunmountmountable.h>
+#include <gvfsjobstartmountable.h>
+#include <gvfsjobstopmountable.h>
+#include <gvfsjobpollmountable.h>
 #include <gvfsjobmakedirectory.h>
 #include <gvfsjobmakesymlink.h>
 #include <gvfsjobcreatemonitor.h>
-#include <gvfsjobupload.h>
 #include <gvfsjobcopy.h>
 #include <gvfsjobmove.h>
+#include <gvfsjobpush.h>
+#include <gvfsjobpull.h>
 #include <gvfsjobsetattribute.h>
 #include <gvfsjobqueryattributes.h>
 #include <gdbusutils.h>
@@ -65,10 +70,12 @@ struct _GVfsBackendPrivate
 {
   GVfsDaemon *daemon;
   char *object_path;
-  
+
+  gboolean is_mounted;
   char *display_name;
   char *stable_name;
-  char *icon;
+  char **x_content_types;
+  GIcon *icon;
   char *prefered_filename_encoding;
   gboolean user_visible;
   GMountSpec *mount_spec;
@@ -142,7 +149,9 @@ g_vfs_backend_finalize (GObject *object)
   
   g_free (backend->priv->display_name);
   g_free (backend->priv->stable_name);
-  g_free (backend->priv->icon);
+  g_strfreev (backend->priv->x_content_types);
+  if (backend->priv->icon != NULL)
+    g_object_unref (backend->priv->icon);
   g_free (backend->priv->prefered_filename_encoding);
   if (backend->priv->mount_spec)
     g_mount_spec_unref (backend->priv->mount_spec);
@@ -190,7 +199,7 @@ static void
 g_vfs_backend_init (GVfsBackend *backend)
 {
   backend->priv = G_TYPE_INSTANCE_GET_PRIVATE (backend, G_VFS_TYPE_BACKEND, GVfsBackendPrivate);
-  backend->priv->icon = g_strdup ("");
+  backend->priv->icon = NULL;
   backend->priv->prefered_filename_encoding = g_strdup ("");
   backend->priv->display_name = g_strdup ("");
   backend->priv->stable_name = g_strdup ("");
@@ -268,6 +277,11 @@ g_vfs_backend_get_daemon (GVfsBackend *backend)
   return backend->priv->daemon;
 }
 
+gboolean
+g_vfs_backend_is_mounted (GVfsBackend *backend)
+{
+  return backend->priv->is_mounted;
+}
 
 void
 g_vfs_backend_set_display_name (GVfsBackend *backend,
@@ -299,12 +313,43 @@ g_vfs_backend_set_stable_name (GVfsBackend        *backend,
   backend->priv->stable_name = g_strdup (stable_name);
 }
 
+/**
+ * g_vfs_backend_set_x_content_types:
+ * @backend: backend
+ * @x_content_types: the x-content types
+ *
+ * For backends where the x-content type is known ahead of time and
+ * won't change (such as a CDDA audio disc backend), this function
+ * should be called when the backend is constructed with the given
+ * types.
+ *
+ * See the <ulink url="http://www.freedesktop.org/wiki/Specifications/shared-mime-info-spec">shared-mime-info</ulink>
+ * specification for more on x-content types.
+ **/
+void
+g_vfs_backend_set_x_content_types (GVfsBackend        *backend,
+                                   char              **x_content_types)
+{
+  g_strfreev (backend->priv->x_content_types);
+  backend->priv->x_content_types = g_strdupv (x_content_types);
+}
+
 void
 g_vfs_backend_set_icon_name (GVfsBackend *backend,
-			     const char *icon)
+			     const char *icon_name)
 {
-  g_free (backend->priv->icon);
-  backend->priv->icon = g_strdup (icon);
+  if (backend->priv->icon != NULL)
+    g_object_unref (backend->priv->icon);
+  backend->priv->icon = g_themed_icon_new_with_default_fallbacks (icon_name);
+}
+
+void
+g_vfs_backend_set_icon (GVfsBackend *backend,
+                        GIcon       *icon)
+{
+  if (backend->priv->icon != NULL)
+    g_object_unref (backend->priv->icon);
+  backend->priv->icon = g_object_ref (icon);
 }
 
 void
@@ -351,8 +396,14 @@ g_vfs_backend_get_stable_name (GVfsBackend *backend)
   return backend->priv->stable_name;
 }
 
-const char *
-g_vfs_backend_get_icon_name (GVfsBackend *backend)
+char **
+g_vfs_backend_get_x_content_types (GVfsBackend *backend)
+{
+  return backend->priv->x_content_types;
+}
+
+GIcon *
+g_vfs_backend_get_icon (GVfsBackend *backend)
 {
   return backend->priv->icon;
 }
@@ -439,7 +490,7 @@ backend_dbus_handler (DBusConnection  *connection,
 
   job = NULL;
 
-  g_print ("backend_dbus_handler %s:%s\n",
+  g_debug ("backend_dbus_handler %s:%s\n",
 	   dbus_message_get_interface (message),
 	   dbus_message_get_member (message));
   
@@ -451,6 +502,10 @@ backend_dbus_handler (DBusConnection  *connection,
 				   G_VFS_DBUS_MOUNT_INTERFACE,
 				   G_VFS_DBUS_MOUNT_OP_OPEN_FOR_READ))
     job = g_vfs_job_open_for_read_new (connection, message, backend);
+  else if (dbus_message_is_method_call (message,
+				   G_VFS_DBUS_MOUNT_INTERFACE,
+				   G_VFS_DBUS_MOUNT_OP_OPEN_ICON_FOR_READ))
+    job = g_vfs_job_open_icon_for_read_new (connection, message, backend);
   else if (dbus_message_is_method_call (message,
 					G_VFS_DBUS_MOUNT_INTERFACE,
 					G_VFS_DBUS_MOUNT_OP_QUERY_INFO))
@@ -479,6 +534,18 @@ backend_dbus_handler (DBusConnection  *connection,
 					G_VFS_DBUS_MOUNT_INTERFACE,
 					G_VFS_DBUS_MOUNT_OP_EJECT_MOUNTABLE))
     job = g_vfs_job_unmount_mountable_new (connection, message, backend, TRUE);
+  else if (dbus_message_is_method_call (message,
+					G_VFS_DBUS_MOUNT_INTERFACE,
+					G_VFS_DBUS_MOUNT_OP_START_MOUNTABLE))
+    job = g_vfs_job_start_mountable_new (connection, message, backend);
+  else if (dbus_message_is_method_call (message,
+					G_VFS_DBUS_MOUNT_INTERFACE,
+					G_VFS_DBUS_MOUNT_OP_STOP_MOUNTABLE))
+    job = g_vfs_job_stop_mountable_new (connection, message, backend);
+  else if (dbus_message_is_method_call (message,
+					G_VFS_DBUS_MOUNT_INTERFACE,
+					G_VFS_DBUS_MOUNT_OP_POLL_MOUNTABLE))
+    job = g_vfs_job_poll_mountable_new (connection, message, backend);
   else if (dbus_message_is_method_call (message,
 					G_VFS_DBUS_MOUNT_INTERFACE,
 					G_VFS_DBUS_MOUNT_OP_SET_DISPLAY_NAME))
@@ -505,8 +572,12 @@ backend_dbus_handler (DBusConnection  *connection,
     job = g_vfs_job_copy_new (connection, message, backend);
   else if (dbus_message_is_method_call (message,
 					G_VFS_DBUS_MOUNT_INTERFACE,
-					G_VFS_DBUS_MOUNT_OP_UPLOAD))
-    job = g_vfs_job_upload_new (connection, message, backend);
+					G_VFS_DBUS_MOUNT_OP_PUSH))
+    job = g_vfs_job_push_new (connection, message, backend);
+  else if (dbus_message_is_method_call (message,
+					G_VFS_DBUS_MOUNT_INTERFACE,
+					G_VFS_DBUS_MOUNT_OP_PULL))
+    job = g_vfs_job_pull_new (connection, message, backend);
   else if (dbus_message_is_method_call (message,
 					G_VFS_DBUS_MOUNT_INTERFACE,
 					G_VFS_DBUS_MOUNT_OP_MOVE))
@@ -551,7 +622,21 @@ g_vfs_backend_register_mount (GVfsBackend *backend,
   DBusMessage *message;
   DBusMessageIter iter;
   dbus_bool_t user_visible;
-  
+  char *x_content_types_string;
+  char *icon_str;
+
+  backend->priv->is_mounted = TRUE;
+
+  if (backend->priv->x_content_types != NULL && g_strv_length (backend->priv->x_content_types) > 0)
+    x_content_types_string = g_strjoinv (" ", backend->priv->x_content_types);
+  else
+    x_content_types_string = g_strdup ("");
+
+  if (backend->priv->icon != NULL)
+    icon_str = g_icon_to_string (backend->priv->icon);
+  else
+    icon_str = g_strdup ("");
+
   message = dbus_message_new_method_call (G_VFS_DBUS_DAEMON_NAME,
 					  G_VFS_DBUS_MOUNTTRACKER_PATH,
 					  G_VFS_DBUS_MOUNTTRACKER_INTERFACE,
@@ -570,7 +655,8 @@ g_vfs_backend_register_mount (GVfsBackend *backend,
 				 DBUS_TYPE_OBJECT_PATH, &backend->priv->object_path,
 				 DBUS_TYPE_STRING, &backend->priv->display_name,
 				 DBUS_TYPE_STRING, &stable_name,
-				 DBUS_TYPE_STRING, &backend->priv->icon,
+				 DBUS_TYPE_STRING, &x_content_types_string,
+				 DBUS_TYPE_STRING, &icon_str,
 				 DBUS_TYPE_STRING, &backend->priv->prefered_filename_encoding,
 				 DBUS_TYPE_BOOLEAN, &user_visible,
 				 0))
@@ -584,6 +670,9 @@ g_vfs_backend_register_mount (GVfsBackend *backend,
   _g_dbus_connection_call_async (NULL, message, -1, 
 				 callback, user_data);
   dbus_message_unref (message);
+
+  g_free (x_content_types_string);
+  g_free (icon_str);
 }
 
 void
@@ -609,3 +698,248 @@ g_vfs_backend_unregister_mount (GVfsBackend *backend,
 				 callback, user_data);
   dbus_message_unref (message);
 }
+
+/* ------------------------------------------------------------------------------------------------- */
+
+typedef struct
+{
+  GVfsBackend *backend;
+  GMountSource *mount_source;
+
+  gboolean ret;
+  gboolean aborted;
+  gint choice;
+
+  const gchar *message;
+  const gchar *choices[3];
+
+  gboolean no_more_processes;
+
+  GAsyncReadyCallback callback;
+  gpointer            user_data;
+
+  guint timeout_id;
+} UnmountWithOpData;
+
+static void
+complete_unmount_with_op (UnmountWithOpData *data)
+{
+  gboolean ret;
+  GSimpleAsyncResult *simple;
+
+  g_source_remove (data->timeout_id);
+
+  ret = TRUE;
+
+  if (data->no_more_processes)
+    {
+      /* do nothing, e.g. return TRUE to signal we should unmount */
+    }
+  else
+    {
+      if (data->aborted || data->choice == 1)
+        {
+          ret = FALSE;
+        }
+    }
+
+  simple = g_simple_async_result_new (G_OBJECT (data->backend),
+                                      data->callback,
+                                      data->user_data,
+                                      NULL);
+  g_simple_async_result_set_op_res_gboolean (simple, ret);
+  g_simple_async_result_complete (simple);
+  g_object_unref (simple);
+}
+
+static void
+on_show_processes_reply (GMountSource  *mount_source,
+                         GAsyncResult  *res,
+                         gpointer       user_data)
+{
+  UnmountWithOpData *data = user_data;
+
+  /* Do nothing if we've handled this already */
+  if (data->no_more_processes)
+    return;
+
+  data->ret = g_mount_source_show_processes_finish (mount_source,
+                                                    res,
+                                                    &data->aborted,
+                                                    &data->choice);
+
+  complete_unmount_with_op (data);
+}
+
+static gboolean
+on_update_processes_timeout (gpointer user_data)
+{
+  UnmountWithOpData *data = user_data;
+  GArray *processes;
+
+  processes = g_vfs_daemon_get_blocking_processes (g_vfs_backend_get_daemon (data->backend));
+  if (processes->len == 0)
+    {
+      /* no more processes, abort mount op */
+      g_mount_source_abort (data->mount_source);
+      data->no_more_processes = TRUE;
+      complete_unmount_with_op (data);
+    }
+  else
+    {
+      /* ignore reply */
+      g_mount_source_show_processes_async (data->mount_source,
+                                           data->message,
+                                           processes,
+                                           data->choices,
+                                           g_strv_length ((gchar **) data->choices),
+                                           NULL,
+                                           NULL);
+    }
+
+  g_array_unref (processes);
+
+  /* keep timeout around */
+  return TRUE;
+}
+
+static void
+unmount_with_op_data_free (UnmountWithOpData *data)
+{
+  g_free (data);
+}
+
+
+/**
+ * g_vfs_backend_unmount_with_operation_finish:
+ * @backend: A #GVfsBackend.
+ * @res: A #GAsyncResult obtained from the @callback function passed
+ *     to g_vfs_backend_unmount_with_operation().
+ *
+ * Gets the result of the operation started by
+ * gvfs_backend_unmount_with_operation_sync().
+ *
+ * Returns: %TRUE if the backend should be unmounted (either no blocking
+ *     processes or the user decided to unmount anyway), %FALSE if
+ *     no action should be taken.
+ */
+gboolean
+g_vfs_backend_unmount_with_operation_finish (GVfsBackend *backend,
+                                             GAsyncResult *res)
+{
+  gboolean ret;
+  GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT (res);
+
+  if (g_simple_async_result_propagate_error (simple, NULL))
+    {
+      ret = FALSE;
+    }
+  else
+    {
+      ret = g_simple_async_result_get_op_res_gboolean (simple);
+    }
+
+  return ret;
+}
+
+/**
+ * gvfs_backend_unmount_with_operation:
+ * @backend: A #GVfsBackend.
+ * @callback: A #GAsyncReadyCallback.
+ * @user_data: User data to pass to @callback.
+ *
+ * Utility function to checks if there are pending operations on
+ * @backend preventing unmount. If not, then @callback is invoked
+ * immediately.
+ *
+ * Otherwise, a dialog will be shown (using @mount_source) to interact
+ * with the user about blocking processes (e.g. using the
+ * #GMountOperation::show-processes signal). The list of blocking
+ * processes is continuously updated.
+ *
+ * Once the user has decided (or if it's not possible to interact with
+ * the user), @callback will be invoked. You can then call
+ * g_vfs_backend_unmount_with_operation_finish() to get the result
+ * of the operation.
+ */
+void
+g_vfs_backend_unmount_with_operation (GVfsBackend        *backend,
+                                      GMountSource       *mount_source,
+                                      GAsyncReadyCallback callback,
+                                      gpointer            user_data)
+{
+  GArray *processes;
+  UnmountWithOpData *data;
+
+  g_return_if_fail (G_VFS_IS_BACKEND (backend));
+  g_return_if_fail (G_IS_MOUNT_SOURCE (mount_source));
+  g_return_if_fail (callback != NULL);
+
+  processes = g_vfs_daemon_get_blocking_processes (g_vfs_backend_get_daemon (backend));
+  /* if no processes are blocking, complete immediately */
+  if (processes->len == 0)
+    {
+      GSimpleAsyncResult *simple;
+      simple = g_simple_async_result_new (G_OBJECT (backend),
+                                          callback,
+                                          user_data,
+                                          NULL);
+      g_simple_async_result_set_op_res_gboolean (simple, TRUE);
+      g_simple_async_result_complete (simple);
+      g_object_unref (simple);
+      goto out;
+    }
+
+  data = g_new0 (UnmountWithOpData, 1);
+  data->backend = backend;
+  data->mount_source = mount_source;
+  data->callback = callback;
+  data->user_data = user_data;
+
+  data->choices[0] = _("Unmount Anyway");
+  data->choices[1] = _("Cancel");
+  data->choices[2] = NULL;
+  data->message = _("Volume is busy\n"
+                    "One or more applications are keeping the volume busy.");
+
+  /* free data when the mount source goes away */
+  g_object_set_data_full (G_OBJECT (mount_source),
+                          "unmount-op-data",
+                          data,
+                          (GDestroyNotify) unmount_with_op_data_free);
+
+  /* show processes */
+  g_mount_source_show_processes_async (mount_source,
+                                       data->message,
+                                       processes,
+                                       data->choices,
+                                       g_strv_length ((gchar **) data->choices),
+                                       (GAsyncReadyCallback) on_show_processes_reply,
+                                       data);
+
+  /* update these processes every two secs */
+  data->timeout_id = g_timeout_add_seconds (2,
+                                            on_update_processes_timeout,
+                                            data);
+
+ out:
+  g_array_unref (processes);
+
+}
+
+gboolean
+g_vfs_backend_has_blocking_processes (GVfsBackend  *backend)
+{
+  gboolean ret;
+  GArray *processes;
+
+  ret = FALSE;
+  processes = g_vfs_daemon_get_blocking_processes (g_vfs_backend_get_daemon (backend));
+  if (processes->len > 0)
+    ret = TRUE;
+
+  g_array_unref (processes);
+
+  return ret;
+}
+
