@@ -99,6 +99,7 @@ gvfs_backend_ftp_determine_features (GVfsFtpTask *task)
     { "MDTM", G_VFS_FTP_FEATURE_MDTM },
     { "SIZE", G_VFS_FTP_FEATURE_SIZE },
     { "TVFS", G_VFS_FTP_FEATURE_TVFS },
+    { "EPRT", G_VFS_FTP_FEATURE_EPRT },
     { "EPSV", G_VFS_FTP_FEATURE_EPSV },
     { "UTF8", G_VFS_FTP_FEATURE_UTF8 },
   };
@@ -527,7 +528,9 @@ try_mount (GVfsBackend *backend,
 
 static void
 do_unmount (GVfsBackend *   backend,
-            GVfsJobUnmount *job)
+            GVfsJobUnmount *job,
+            GMountUnmountFlags flags,
+            GMountSource *mount_source)
 {
   GVfsBackendFtp *ftp = G_VFS_BACKEND_FTP (backend);
   GVfsFtpConnection *conn;
@@ -642,7 +645,7 @@ do_open_for_read (GVfsBackend *backend,
                                                          error_550_permission_or_not_found, 
                                                          NULL };
 
-  g_vfs_ftp_task_open_data_connection (&task);
+  g_vfs_ftp_task_setup_data_connection (&task);
   file = g_vfs_ftp_file_new_from_gvfs (ftp, filename);
 
   g_vfs_ftp_task_send_and_check (&task,
@@ -653,6 +656,8 @@ do_open_for_read (GVfsBackend *backend,
                                  "RETR %s", g_vfs_ftp_file_get_ftp_path (file));
   g_vfs_ftp_file_free (file);
 
+  g_vfs_ftp_task_open_data_connection (&task);
+
   if (!g_vfs_ftp_task_is_in_error (&task))
     {
       /* don't push the connection back, it's our handle now */
@@ -722,7 +727,7 @@ do_start_write (GVfsFtpTask *task,
 
   /* FIXME: can we honour the flags? */
 
-  g_vfs_ftp_task_open_data_connection (task);
+  g_vfs_ftp_task_setup_data_connection (task);
 
   va_start (varargs, format);
   g_vfs_ftp_task_sendv (task,
@@ -732,6 +737,8 @@ do_start_write (GVfsFtpTask *task,
                         varargs);
   va_end (varargs);
 
+  g_vfs_ftp_task_open_data_connection (task);
+
   if (!g_vfs_ftp_task_is_in_error (task))
     {
       /* don't push the connection back, it's our handle now */
@@ -1280,13 +1287,14 @@ do_pull (GVfsBackend *         backend,
     }
   else
     total_size = 0;
-  g_vfs_ftp_task_open_data_connection (&task);
+  g_vfs_ftp_task_setup_data_connection (&task);
   g_vfs_ftp_task_send_and_check (&task,
                                  G_VFS_FTP_PASS_100 | G_VFS_FTP_FAIL_200,
                                  &open_read_handlers[0],
                                  src,
                                  NULL,
                                  "RETR %s", g_vfs_ftp_file_get_ftp_path (src));
+  g_vfs_ftp_task_open_data_connection (&task);
   if (g_vfs_ftp_task_is_in_error (&task))
     {
       g_vfs_ftp_file_free (src);
